"use strict";

exports.__esModule = true;
exports.Tag = exports.TypeInfo = exports.newSymbol = exports.symName = undefined;
exports.symInfo = symInfo;
exports.typeInfo = typeInfo;
exports.nodeInfo = nodeInfo;
exports.symKind = symKind;
exports.fieldInfo = fieldInfo;
exports.symbol = symbol;
exports.enter = enter;
exports.leave = leave;
exports.tok = tok;
exports.produce = produce;
exports.toArray = toArray;
exports.consume = consume;
exports.reproduceNodes = reproduceNodes;
exports.resetFieldInfo = resetFieldInfo;
exports.removeNulls = removeNulls;

var _babelTypes = require("babel-types");

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var GLOBAL_SYMBOLS = true;

var nameCount = 0;
var symbols = new Map();
//TODO: fields used only with a single type
var fieldTypes = new Map();

function symInfo(sym) {
  return symbols.get(sym);
}

function typeInfo(i) {
  return symInfo(i.type);
}

var symName = exports.symName = GLOBAL_SYMBOLS ? Symbol.keyFor : s => symInfo(s).name;

var newSymbol = exports.newSymbol = GLOBAL_SYMBOLS ? Symbol.for : Symbol;

function nodeInfo(node) {
  return symInfo(Tag[node.type]);
}

function symKind(sym) {
  return symInfo(sym).kind;
}

function fieldInfo(type, field) {
  var e = symInfo(type);
  if (e == null || e.fieldsMap == null) return null;
  return e.fieldsMap.get(field);
}

function symbol(name) {
  var kind = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : "ctrl";

  assert.ok(name && name.substr);
  assert.ok(isNaN(name));
  var res = newSymbol(name);
  symbols.set(res, {
    sym: res,
    name,
    kind,
    x: nameCount++,
    prop: null
  });
  return res;
}

function symbolDefFor(name, kind) {
  assert.ok(name && name.substr);
  assert.ok(isNaN(name));
  var sym = Tag[name],
      def = void 0;
  if (sym == null) {
    Tag[name] = sym = newSymbol(name);
    def = { sym, name, kind, x: nameCount++, expr: false,
      block: false,
      key: false,
      lval: false, decl: false, func: false };
    if (kind === "node") def.esType = name;
    symbols.set(sym, def);
  } else {
    def = symInfo(sym);
    assert.ok(def);
    assert.equal(kind, def.kind, `for ${name}`);
  }
  return def;
}

var TypeInfo = exports.TypeInfo = {};

var Tag = exports.Tag = { push: symbol("push", "pos"),
  top: symbol("top", "pos"),
  Array: symbol("Array", "array"),
  Node: symbol("Node", "node"),
  Null: symbol("Null", "null") };

for (var i in _babelTypes.VISITOR_KEYS) {
  var def = TypeInfo[i] = symbolDefFor(i, "node");
  for (var j of _babelTypes.VISITOR_KEYS[i]) {
    symbolDefFor(j, "pos").visitor = true;
  }for (var _j of _babelTypes.BUILDER_KEYS[i]) {
    symbolDefFor(_j, "pos").builder = true;
  }
}
for (var _i in _babelTypes.ALIAS_KEYS) {
  var _def = symbolDefFor(_i, "node");
  var aliases = _def.aliases || (_def.aliases = new Set());
  var aliasKeys = _babelTypes.ALIAS_KEYS[_i];
  if (aliasKeys != null) {
    for (var _j2 of aliasKeys) {
      var adef = symbolDefFor(_j2, "alias");
      (adef.instances || (adef.instances = new Set())).add(_def.sym);
      aliases.add(adef.sym);
    }
  }
}
{
  var idDef = symInfo(Tag.Identifier);
  idDef.fieldsMap = new Map([[Tag.name, { atomicType: "string",
    nodeTypes: new Set(),
    nillable: false,
    enumValues: null,
    default: null
  }]]);
  var meDef = symInfo(Tag.MemberExpression);
  meDef.fieldsMap = new Map([[Tag.property, { atimicType: null,
    nodeTypes: new Set([Tag.Identifier]),
    nillable: false,
    enumValues: null,
    default: null }]]);
}

/** Tag.Identifier not declaring any reference */
var notDeclsPosSet = new Set([Tag.label, Tag.key, Tag.imported, Tag.exported, Tag.label, Tag.key, Tag.meta, Tag.property]);

var declsPosSet = new Set([Tag.id]);

var _loop = function (_i2) {
  var nodeFields = _babelTypes.NODE_FIELDS[_i2];
  var pos = Tag[_i2];
  var def = symInfo(pos);
  if (nodeFields != null) {
    var fieldsMap = def.fieldsMap || (def.fieldsMap = new Map());
    for (var _j3 in nodeFields) {
      var jdef = nodeFields[_j3];
      var fdef = symbolDefFor(_j3, "pos");
      if (fieldsMap.has(fdef.sym)) continue;
      var info = getTy(jdef.validate, fdef.sym);
      info.default = jdef.default;
      fieldsMap.set(fdef.sym, info);
    }
  }
  function getTy(ty, pos) {
    var enumValues = null,
        nt = new Set(),
        atomicType = null,
        nillable = false;
    if (ty != null && ty.chainOf != null) {
      assert.equal(ty.chainOf.length, 2);
      if (ty.chainOf[0].type === "array") {
        var elem = getTy(ty.chainOf[1].each);
        if (pos === Tag.params) elem.declVar = true;
        return {
          array: true,
          elem: elem,
          fieldsMap: new Map([[Tag.push, elem]])
        };
      } else if (ty.chainOf[0].type === "string") {
        assert.ok(ty.chainOf[1].oneOf);
        enumValues = ty.chainOf[1].oneOf;
        atomicType = "string";
        ty = null;
      } else {
        throw new Error("not implemented");
      }
    }
    if (ty != null) {
      if (ty.type != null) {
        atomicType = ty.type;
      } else if (ty.oneOfNodeTypes != null) {
        for (var k of ty.oneOfNodeTypes) {
          var p = Tag[k];
          assert.ok(p);
          nt.add(p);
        }
      } else if (ty.oneOf != null) {
        enumValues = ty.oneOf;
        atomicType = "string";
      } else if (ty.oneOfNodeOrValueTypes != null) {
        nt = new Set();
        for (var _k of ty.oneOfNodeOrValueTypes) {
          if (_k === "null") {
            nillable = true;
          } else {
            var _p = Tag[_k];
            assert.ok(_p, `no such type ${_k}, ${Object.keys(Tag)}`);
            nt.add(_p);
          }
        }
      } else {
        atomicType = "any";
      }
    }
    if (enumValues != null) {
      nillable = nillable || enumValues.indexOf(null) !== -1;
      enumValues = enumValues.filter(i => i != null);
      assert.equal(enumValues.filter(v => v.substr == null).length, 0);
    }
    var expr = nt.has(Tag.Expression),
        lval = nt.has(Tag.LVal);
    return {
      atomicType, nodeTypes: nt, nillable, enumValues, expr,
      stmt: nt.has(Tag.Statement),
      block: nt.has(Tag.BlockStatement) && !nt.has(Tag.Statement),
      key: nt.has(Tag.Identifier) && !expr && !lval,
      lval,
      decl: nt.has(Tag.VariableDeclaration) || nt.has(Tag.Declaration),
      mod: lval,
      declVar: !expr && (nt.has(Tag.Identifier) && !notDeclsPosSet.has(pos) || declsPosSet.has(pos))
    };
  }
};

for (var _i2 in _babelTypes.VISITOR_KEYS) {
  _loop(_i2);
}

for (var _i3 in _babelTypes.VISITOR_KEYS) {
  var _def2 = symInfo(Tag[_i3]);
  var _aliases = _def2.aliases;
  _def2.func = _aliases.has(Tag.Function);
  _def2.scope = _aliases.has(Tag.FunctionParent);
  _def2.expr = _aliases.has(Tag.Expression);
  _def2.decl = _aliases.has(Tag.Declaration);
  _def2.stmt = _aliases.has(Tag.BlockStatement) || _aliases.has(Tag.Statement);
  _def2.block = _aliases.has(Tag.BlockStatement) && !_aliases.has(Tag.Statement);
}

for (var _i4 in Tag) {
  Tag[Tag[_i4]] = Tag[_i4];
}

function setComputed(sym, prop, tys) {
  var me = symInfo(sym);
  me.propAlt = Object.assign({}, me, { fieldsMap: new Map(me.fieldsMap).set(prop, { atomicType: null,
      nillable: false,
      enumValues: null,
      expr: true,
      stmt: false,
      block: false,
      lval: false,
      decl: false,
      mod: false,
      declVar: false,
      nodeTypes: new Set([Tag.Expression]) }) });
  me.prop = Tag.computed;
}

setComputed(Tag.MemberExpression, Tag.property, [Tag.Identifier]);
setComputed(Tag.ObjectProperty, Tag.key, [Tag.Identifier, Tag.StringLiteral, Tag.NumericLiteral]);
setComputed(Tag.ObjectMethod, Tag.key, [Tag.Identifier, Tag.StringLiteral, Tag.NumericLiteral]);
symInfo(Tag.UpdateExpression).fieldsMap.get(Tag.argument).mod = true;
symInfo(Tag.BlockStatement).block = true;
{
  var me = symInfo(Tag.AssignmentExpression);
  me.propAlt = Object.assign({}, me, { fieldsMap: new Map(me.fieldsMap).set(Tag.left, Object.assign({}, me.fieldsMap.get(Tag.left), { expr: true })) });
  me.prop = Tag.operator;
}
{
  // fixing patterns
  var patField = symInfo(Tag.VariableDeclarator).fieldsMap.get(Tag.id);
  var oprop = symInfo(Tag.ObjectProperty);
  var aprop = symbolDefFor("AssignmentProperty", "node");
  aprop.esType = "ObjectProperty";
  aprop.fieldsMap = new Map(oprop.fieldsMap);
  aprop.fieldsMap.set(Tag.value, patField);
  var op = symInfo(Tag.ObjectPattern);
  var el = op.fieldsMap.get(Tag.properties).elem;
  el.decl = el.declVar = true;
  el.ti = aprop;
  var els = symInfo(Tag.ArrayPattern).fieldsMap.get(Tag.elements);
  els.elem = patField;
  els.fieldsMap.set(Tag.push, patField);
  symInfo(Tag.RestElement).fieldsMap.set(Tag.argument, patField);
}

function isNode(node) {
  if (node == null) return false;
  return typeof node === 'object' && typeof node.type === 'string';
}

function enter(pos, type, value) {
  return { enter: true, leave: false, pos, type, value };
}

function leave(pos, type, value) {
  return { enter: false, leave: true, pos, type, value };
}

function tok(pos, type, value) {
  return { enter: true, leave: true, pos, type, value };
}

function* produce(node, pos) {
  function* walk(node, pos) {
    if (Array.isArray(node)) {
      var value = { node };
      yield enter(pos, Tag.Array, value);
      for (var _i5 of node) {
        yield* walk(_i5, Tag.push);
      }yield leave(pos, Tag.Array, value);
    } else if (isNode(node)) {
      var keys = _babelTypes.VISITOR_KEYS[node.type];
      var ti = nodeInfo(node);
      var type = ti.sym;
      if (keys.length) {
        var _value = { node };
        yield enter(pos, type, _value);
        for (var _i6 of keys) {
          var v = node[_i6];
          if (v != null) yield* walk(node[_i6], Tag[_i6] || _i6);
        }
        yield leave(pos, type, _value);
      } else {
        yield tok(pos, type, { node });
      }
    }
  }
  yield* walk(node, pos || Tag.top);
}

function toArray(s) {
  if (Array.isArray(s)) return s;
  return [].concat(_toConsumableArray(s));
}

/**
 * same as consume but returns token's sequence with all fields
 * reset in its values
 */
function* reproduce(s) {
  var stack = [{}];
  var res = [];
  for (var _i7 of s) {
    res.push(_i7);
    if (_i7.type == null || !Tag[Symbol.keyFor(_i7.type)]) continue;
    if (_i7.enter) {
      if (_i7.type === Tag.Array) {
        stack.unshift(_i7.value.node = []);
      } else {
        if (_i7.value != null) _i7.value.type = symName(_i7.type);
        stack.unshift(_i7.value.node);
      }
    }
    if (_i7.leave) {
      var value = stack.shift();
      if (_i7.pos === Tag.push) {
        stack[0].push(value);
      } else stack[0][Symbol.keyFor(_i7.pos)] = value;
    }
  }
  return res;
}

function consume(s) {
  var stack = [{}];
  for (var _i8 of s) {
    var ti = typeInfo(_i8);
    if (_i8.type == null || ti.kind === "ctrl") continue;
    if (_i8.enter) {
      if (_i8.type === Tag.Array) stack.unshift([]);
      //TODO: another step to handle nulls
      else if (_i8.type === Tag.Null) {
          if (_i8.pos !== Tag.push) stack[0][symName(_i8.pos)] = null;
          continue;
        } else {
          if (_i8.value != null) {
            if (ti.esType != null) _i8.value.node.type = ti.esType;
            if (ti.fields) Object.assign(_i8.value.node, ti.fields);
          }
          assert.ok(_i8.value.node.type);
          stack.unshift(_i8.value.node);
        }
    }
    if (_i8.leave) {
      var node = stack.shift();
      if (_i8.pos === Tag.push) {
        stack[0].push(node);
      } else stack[0][symName(_i8.pos)] = node;
    }
  }
  return stack[0];
}

/** 
 * unwraps tokens formerly folded into Node sub-token
 */
function* reproduceNodes(s) {
  for (var _i9 of s) {
    if (_i9.type === Tag.Node) {
      if (_i9.enter) {
        yield* produce(_i9.value.node, _i9.pos);
      } else yield _i9;
    }
  }
}

function* resetFieldInfo(s) {
  var stack = [];
  for (var _i10 of s) {
    if (_i10.enter) {
      var f = stack[stack.length - 1];
      if (f && f.fieldsMap) f = _i10.value.fieldInfo = f.fieldsMap.get(_i10.pos);
      var ti = f && f.ti || typeInfo(_i10);
      switch (ti.kind) {
        case "array":
          stack.push(_i10.value.fieldInfo);
          break;
        case "node":
          if (ti.prop != null && _i10.value.node != null) {
            if (ti.prop === Tag.operator) {
              if (_i10.value.node.operator !== "=") ti = ti.propAlt;
            } else {
              if (_i10.value.node.computed) ti = ti.propAlt;
            }
          }
          stack.push(ti);
          break;
        default:
          stack.push(false);
      }
    }
    if (_i10.leave) stack.pop();
    yield _i10;
  }
}

function* removeNulls(s) {
  var stack = [];
  for (var _i11 of s) {
    if (_i11.type === Tag.Null) {
      if (_i11.enter && stack[0]) {
        if (_i11.pos != Tag.push) stack[0][symName(_i11.pos)] = null;
      }
      continue;
    }
    yield _i11;
    if (_i11.enter) stack.unshift(_i11.value.node);
    if (_i11.leave) stack.shift();
  }
}