"use strict";

exports.__esModule = true;
exports.fin = exports.toConsole = exports.color = exports.convertCtrl = undefined;
exports.markNodeType = markNodeType;
exports.toStr = toStr;
exports.setComment = setComment;
exports.setEndComment = setEndComment;
exports.copyComment = copyComment;
exports.cleanComments = cleanComments;

var _ramda = require("ramda");

var R = _interopRequireWildcard(_ramda);

var _kit = require("./kit");

var Kit = _interopRequireWildcard(_kit);

var _trace = require("./trace");

var Trace = _interopRequireWildcard(_trace);

var _core = require("./core");

var _fs = require("fs");

var fs = _interopRequireWildcard(_fs);

var _babelGenerator = require("babel-generator");

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

var _babelTypes = require("babel-types");

var T = _interopRequireWildcard(_babelTypes);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var BROWSER_DEBUG = Trace.BROWSER_DEBUG;

function* markNodeType(s) {
  for (var i of s) {
    if (i.enter) {
      var ti = (0, _core.typeInfo)(i);
      if (ti.kind === "ctrl") (i.value.comments || (i.value.comments = [])).unshift({ txt: (0, _core.symName)(i.type), style: styles.nodetype });else if (i.value.comments != null && i.value.comments.length) {
        (i.value.comments || (i.value.comments = [])).unshift({ txt: (0, _core.symName)(i.type).match(/[A-Z]/g).join(""),
          style: styles.nodetype });
      }
    }
    yield i;
  }
}

var convertCtrl = exports.convertCtrl = R.pipe(_core.resetFieldInfo, markNodeType, Array.from, function convertCtrl(s) {
  s = Kit.auto(s);
  function* walk(sw) {
    for (var i of sw) {
      if (i.enter) {
        var fld = i.value.fieldInfo || {},
            ti = (0, _core.typeInfo)(i);
        if (ti.kind === "ctrl") {
          s.peel(i);
          var lab = s.label();
          var nm = ti.name;
          var stmt = !fld.expr;
          yield s.enter(i.pos, stmt ? _core.Tag.BlockStatement : _core.Tag.SequenceExpression, { comments: i.value.comments });
          yield s.enter(stmt ? _core.Tag.body : _core.Tag.expressions, _core.Tag.Array);
          if (!i.leave) {
            for (var j; (j = s.curLev()) != null;) {
              yield s.enter(_core.Tag.push, Kit.Subst);
              yield* walk(s.one());
              yield* s.leave();
            }
          }
          yield* lab();
          Kit.skip(s.leave());
          continue;
        }
      }
      yield i;
    }
  }
  return walk(s);
}, Kit.completeSubst, Array.from, Kit.adjustFieldType);

var color = exports.color = BROWSER_DEBUG ? function* color(s) {
  for (var i of s) {
    if (i.enter && i.value.comments) {
      for (var j of i.value.comments) {
        if (j.style != null && (0, _core.typeInfo)(i).kind === "node") {
          j.txt = `%c${j.txt}%c`;
          j.args = [j.style, ""];
        }
      }
    }
    if (i.leave && i.value.tcomments) {
      for (var _j of i.value.tcomments) {
        if (_j.style != null && (0, _core.typeInfo)(i).kind === "node") {
          _j.txt = `%c${_j.txt}%c`;
          _j.args = [_j.style, ""];
        }
      }
    }
    yield i;
  }
} : s => s;

function* getArgs(s) {
  for (var i of s) {
    if (i.enter && i.value.comments) {
      for (var j of i.value.comments) {
        if (j.args) for (var k of j.args) {
          yield k;
        }
      }
    }
    if (i.leave && i.value.tcomments) {
      for (var _j2 of i.value.tcomments) {
        if (_j2.args) for (var _k of _j2.args) {
          yield _k;
        }
      }
    }
  }
}

function* applyComments(s) {
  for (var i of s) {
    var node = i.value.node;
    if (i.enter && node != null) {
      if (i.value.comments != null && i.value.comments.length) {
        (node.leadingComments || (node.leadingComments = [])).push({ type: "CommentBlock",
          value: i.value.comments.map(v => v.txt).join("|") });
      }
      if (i.value.tcomments != null && i.value.tcomments.length) {
        (node.trailingComments || (node.trailingComments = [])).push({ type: "CommentBlock",
          value: i.value.tcomments.map(v => v.txt).join("|") });
      }
    }
    yield i;
  }
}

var toConsole = exports.toConsole = R.curry(function toConsole(tag, s) {
  var _console;

  if (BROWSER_DEBUG) console.group(`dump %c${tag}`, "color:orange;font-size:large");else console.log(`dump ${tag}`);
  var col = R.pipe(convertCtrl, color, Array.from, applyComments, Array.from)(s);
  var args = Array.from(getArgs(col));
  (0, _core.consume)(Trace.verify(col));
  (_console = console).log.apply(_console, [(0, _babelGenerator2.default)(col[0].value.node).code].concat(_toConsumableArray(args)));
  if (BROWSER_DEBUG) console.groupEnd();
});

var fin = exports.fin = R.pipe(_core.removeNulls, convertCtrl, Array.from, applyComments, Array.from);

function toStr(s) {
  (0, _core.consume)(fin(s));
  return (0, _babelGenerator2.default)(fin[0].value.node).code;
}

exports.default = R.curry(function dump(tag, s) {
  var sa = Kit.toArray(s);
  var sl = Kit.auto(Kit.clone(sa));
  var opts = sl.opts || {};
  var dest = void 0;
  if (BROWSER_DEBUG) {
    dest = "console";
  } else {
    dest = `dump-${tag}.js`;
    if (opts.dump) {
      if (opts.dump.substr) {
        if (opts.dump === "console") {
          dest = opts.dump;
        } else {
          dest = opts.dump + dest;
        }
      } else {
        var _s = opts.dump[tag];
        if (_s == null && _s.substr) {
          dest = _s;
        }
      }
    }
  }
  if (dest === "console") {
    toConsole(tag, sl);
  } else {
    fs.writeFileSync(dest, toStr(sl));
  }
  return sa;
});
function setComment(i, txt) {
  var style = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "none";

  style = style.substr && styles[style] || style;
  (i.value.comments || (i.value.comments = [])).push({ txt, style });
  return i;
}

function setEndComment(i, txt) {
  var style = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : "none";

  style = style.substr && styles[style] || style;
  (i.value.tcomments || (i.value.tcomments = [])).push({ txt, style });
  return i;
}

function copyComment(f, t) {
  var _ref, _ref2;

  if (f.value.comments != null) (_ref = t.value.comments || (t.value.comments = [])).push.apply(_ref, _toConsumableArray(f.value.comments));
  if (f.value.tcomments != null) (_ref2 = t.value.tcomments || (t.value.tcomments = [])).push.apply(_ref2, _toConsumableArray(f.value.tcomments));
  return t;
}

function* cleanComments(s) {
  for (var i of s) {
    yield i;
    if (i.leave) {
      i.value.comments = null;
      i.value.tcomments = null;
      if (i.value.node) {
        if (i.value.leadingComments) i.value.leadingComments = null;
        if (i.value.trailingComments) i.value.trailingComments = null;
      }
    }
  }
}

var styles = {
  large: "font-size:xx-large;color:orange",
  small: "font-size:large;color:navy;",
  nodetype: "font-size:xx-small;color:green;font-weight:bolder",
  none: ""
};