"use strict";

exports.__esModule = true;
exports.run = exports.inject = exports.clean = exports.commit = exports.Placeholder = exports.Root = undefined;

var _kit = require("./kit");

var Kit = _interopRequireWildcard(_kit);

var _ramda = require("ramda");

var R = _interopRequireWildcard(_ramda);

var _core = require("./core");

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

var _trace = require("./trace");

var Trace = _interopRequireWildcard(_trace);

var _dump = require("./dump");

var _dump2 = _interopRequireDefault(_dump);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

var Root = exports.Root = (0, _core.symbol)("match.root", "ctrl");
var Placeholder = exports.Placeholder = (0, _core.symbol)("match.placeholder", "ctrl");

var commit = exports.commit = R.pipe(Array.from, function* (s) {
  var buf = [];
  for (var i of s) {
    switch (i.type) {
      case Root:
        if (!i.value.v.match) continue;
        i = i.value.s ? (0, _core.enter)(i.pos, i.type, i.value.v) : (0, _core.leave)(i.pos, i.type, i.value.v);
        break;
      case Placeholder:
        if (!i.value.v.match) continue;
    }
    yield i;
  }
});

var clean = exports.clean = function* (s) {
  for (var i of s) {
    switch (i.type) {
      case Root:
      case Placeholder:
        break;
      default:
        yield i;
    }
  }
};

var inject = exports.inject = R.curry(function* matchInject(pattern, si) {
  var pats = (Array.isArray(pattern) ? pattern : [pattern]).map(i => Kit.toArray(Kit.toks(_core.Tag.top, i)));
  var starts = pats.map((i, x) => i[0].type);
  var plen = pats.length;
  var plens = pats.map(i => i.length);
  assert.equal(plens.filter(v => v === 0).length, 0);
  var s = Kit.lookahead(si);
  var activePos = [];
  var activeTok = [];
  var activePh = [];
  var activePat = [];
  var level = 0;
  for (var i of s) {
    if (i.enter) {
      level++;
    }
    var aplen = activePos.length;
    for (var p = 0; p < aplen; ++p) {
      var x = activePos[p];
      var pat = activePat[p];
      var _plen = pat.length;
      var v = activeTok[p];
      var ph = activePh[p];
      if (x !== _plen && x !== -1 && activePh[p] == null) {
        var j = pat[x++];
        //during generation we may neglect emitting empty arrays 
        if (j.pos !== i.pos && j.type === _core.Tag.Array && j.enter && !j.leave && pat[x].value === j.value) {
          x++;
          j = pat[x++];
        }
        if (x < _plen && j.pos != i.pos && j.enter === i.enter) {
          activePos[p] = -1;
          v.match = false;
          continue;
        }
        if (j.enter) {
          switch (j.type) {
            case _core.Tag.ExpressionStatement:
              var k = pat[x];
              if (k.type === _core.Tag.Identifier) {
                var _block = false;
                if (k.value.node.name[0] === "$") {
                  var _ph = activePh[p] = { v, level, name: k.value.node.name.substr(1) };
                  yield (0, _core.enter)(i.pos, Placeholder, _ph);
                  x++;
                  assert.equal(pat[x].value, k.value);
                  x++;
                  assert.equal(pat[x].value, j.value);
                  x++;
                  activePos[p] = x;
                  continue;
                }
              }
              break;
            case _core.Tag.Identifier:
            case _core.Tag.StringLiteral:
              var block = false;
              var name = j.value.node.name || j.value.node.value;
              if (name[0] === "$") {
                var _ph2 = activePh[p] = { v, level, name: name.substr(1) };
                yield (0, _core.enter)(i.pos, Placeholder, _ph2);
                if (!j.leave) {
                  assert.equal(pat[x].value, j.value);
                  x++;
                }
                activePos[p] = x;
                continue;
              }
              break;
          }
        }
        if (j.type !== i.type || j.value.node.name !== i.value.node.name || j.value.node.value !== i.value.node.value) {
          activePos[p] = -1;
          v.match = false;
          continue;
        }
        if (i.enter) {
          assert.ok(j.enter);
          if (i.leave && !j.leave) {
            if (pat[x++].value !== j.value) {
              activePos[p] = -1;
              v.match = false;
              continue;
            }
          } else if (j.leave) {
            if (s.cur().value === i.value) s.read();
          }
        }
        activePos[p] = x;
      }
    }
    if (i.enter) {
      for (var _s = 0; _s < plen; ++_s) {
        if (starts[_s] == i.type) {
          var _v = { match: null, index: _s };
          yield (0, _core.tok)(i.pos, Root, { s: true, v: _v });
          activePos.push(1);
          activeTok.push(_v);
          activePat.push(pats[_s]);
        }
      }
    }
    yield i;
    if (i.leave) {
      var _aplen = activePos.length;
      for (var _p = _aplen - 1; _p >= 0; --_p) {
        var _ph3 = activePh[_p];
        if (_ph3 != null) {
          if (_ph3.level === level) {
            yield (0, _core.leave)(i.pos, Placeholder, _ph3);
            _ph3 = activePh[_p] = null;
          }
        }
        if (_ph3 == null) {
          var _x = activePos[_p];
          if (_x >= 0 && _x === activePat[_p].length) {
            activePos[_p] = -1;
            var _v2 = activeTok[_p];
            _v2.match = true;
            while (activePos[0] === -1) {
              activePos.shift();
              activePh.shift();
              activeTok.shift();
              activePat.shift();
            }
            yield (0, _core.tok)(i.pos, Root, { s: false, v: _v2 });
          }
        }
      }
      level--;
    }
  }
});

var run = exports.run = R.curry(function (pat, si) {
  return commit(inject(pat, si));
});