"use strict";

exports.__esModule = true;

var _core = require("../core");

var _kit = require("../kit");

var Kit = _interopRequireWildcard(_kit);

var _ramda = require("ramda");

var R = _interopRequireWildcard(_ramda);

var _babelTypes = require("babel-types");

var T = _interopRequireWildcard(_babelTypes);

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

//TODO:
function collectFuncDecls(s) {
  var sa = Kit.toArray(s);
  var sl = Kit.levels(sa);
  function walk() {
    for (var i of sl.sub()) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionDefinition:
          case _core.Tag.FunctionExpression:
          case _core.Tag.ArrowFunctionExpression:
        }
      }
    }
  }
  walk(new Set());
  return sa;
}

function markKeyIds(s) {
  var sl = Kit.levels(s);
  function* walk(sw) {
    for (var i of sw) {
      yield i;
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionDeclaration:
          case _core.Tag.FunctionExpression:
            var j = sl.cur();
            if (j.pos === _core.Tag.id) {
              j.value.key = true;
              yield* sl.one();
            }
            break;
          case _core.Tag.MemberExpression:
            yield* walk(sl.one());
            var k = sl.cur();
            if (k.type === _core.Tag.Identifier) {
              k.value.key = true;
              yield* sl.one();
            }
            break;
        }
      }
    }
  }
  return walk(sl);
}

/**
 * joins not computed MemberExpressions into a single variable 
 * if its object is const and not used anywhere else
 */
exports.default = R.pipe(markKeyIds, collectFuncDecls, function joinMemExprCollect(s) {
  var sa = Kit.toArray(s);
  var sl = Kit.levels(sa);
  function walk(dir) {
    for (var i of sl.sub()) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.FunctionDefinition:
          case _core.Tag.FunctionExpression:
          case _core.Tag.ArrowFunctionExpression:
            //TODO:
            break;
          case _core.Tag.ImportNamespaceSpecifier:
            //TODO:
            break;
          case _core.Tag.VariableDeclaration:
            if (i.pos === _core.Tag.push) {
              var vars = i.value.vars = [];
              for (var j of sl.sub()) {
                if (j.enter && j.type === _core.Tag.VariableDeclarator) {
                  for (var k of sl.one()) {
                    if (k.enter && k.type === _core.Tag.Identifier) {
                      switch (i.value.node.kind) {
                        case "const":
                          var pack = Kit.hasAnnot(i.value.node, "@PACK") || Kit.hasAnnot(j.value.node, "@PACK");
                          var info = { name: k.value.node.name,
                            used: new Set(),
                            enabled: true,
                            pack };

                          vars.push(info);
                          dir.set(k.value.node.name, info);
                          break;
                        case "let":
                          dir.delete(k.value.node.name);
                          break;
                      }
                    }
                  }
                  var n = sl.cur();
                  if (n.type === _core.Tag.AssignmentExpression && n.pos === _core.Tag.init) {
                    for (var _j of vars) {
                      _j.enabled = false;
                    }
                  }
                  walk(dir);
                }
              }
            }
            break;
          case _core.Tag.Identifier:
            if (!i.value.key) {
              var _info = dir.get(i.value.node.name);
              if (_info != null && !_info.pack) {
                _info.enabled = false;
              }
            }
            break;
          case _core.Tag.AssignmentExpression:
          case _core.Tag.CallExpression:
            var c = sl.cur();
            if (c.type === _core.Tag.MemberExpression) {
              c.value.discharge = true;
            }
            break;
          case _core.Tag.MemberExpression:
            var o = sl.cur();
            if (o.type === _core.Tag.Identifier && !i.value.node.computed) {
              var _info2 = dir.get(o.value.node.name);
              if (_info2 != null) {
                if (i.value.discharge && !_info2.pack) {
                  _info2.enabled = false;
                  break;
                }
                Kit.skip(sl.one());
                var p = sl.cur();
                if (p.type !== _core.Tag.Identifier && !_info2.pack) {
                  _info2.enabled = false;
                  break;
                }
                Kit.skip(sl.one());
                var path = i.value.path = [o.value.node.name, p.value.node.name];
                i.value.varInfo = _info2;
                _info2.used.add(p.value.node.name);
              }
            }
            break;
          case _core.Tag.BlockStatement:
            walk(new Map(dir));
        }
      }
    }
  }
  walk(new Map());
  return sa;
}, function joinMemExprReplace(s) {
  var sl = Kit.auto(s);
  function* walk() {
    for (var i of sl.sub()) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.MemberExpression:
            if (i.value.varInfo != null && i.value.varInfo.enabled) {
              Kit.skip(sl.copy(i));
              yield sl.tok(i.pos, T.identifier(i.value.path.join("$")));
              continue;
            }
            break;
          case _core.Tag.VariableDeclaration:
            if (i.value.vars != null) {
              yield sl.peel(i);
              yield* walk();
              yield* sl.leave();
              for (var j of i.value.vars) {
                if (j.enabled) {
                  for (var k of j.used) {
                    yield* sl.toks(_core.Tag.push, `let ${j.name}$${k} = ${j.name} && ${j.name}.${k}`);
                  }
                }
              }
              continue;
            }
        }
      }
      yield i;
    }
  }
  return walk();
});