"use strict";

exports.__esModule = true;
exports.resolve = exports.prepare = exports.assignSym = exports.resetSym = undefined;
exports.newSym = newSym;
exports.calcUnorderedScope = calcUnorderedScope;
exports.calcRefScopes = calcRefScopes;
exports.calcBlockRefs = calcBlockRefs;
exports.solve = solve;

var _kit = require("./kit");

var Kit = _interopRequireWildcard(_kit);

var _ramda = require("ramda");

var R = _interopRequireWildcard(_ramda);

var _core = require("./core");

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

var _trace = require("./trace");

var Trace = _interopRequireWildcard(_trace);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var globals = new Map();

var symNum = 0;
var curSymId = 0;

// String -> Sym
function newSym() {
  var name = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : "";

  return { name, orig: name, id: `${name}@${curSymId++}` };
}

/** 
 * resets symbol number and value decl field 
 * for new identifiers from transform passes
 */
var resetSym = exports.resetSym = R.pipe(_core.resetFieldInfo, function* resetSym(s) {
  for (var i of s) {
    if (i.enter && i.type === _core.Tag.Identifier && i.value.sym != null) {
      if (i.value.decl == null) {
        var fi = i.value.fieldInfo;
        i.value.decl = fi.declVar;
      }
      if (i.value.decl) {
        if (i.value.sym.num == null) i.value.sym.num = symNum++;
        i.value.sym.name = i.value.sym.orig;
      }
    }
    yield i;
  }
});

/**
 * calculates unordered definitions scope, i.e, var kind variable declarations
 * or function declarations
 * 
 * for scope nodes:
 *
 *     type Value = Value & {varScope: Map<String,Value>}
 * 
 * the Value in the Map pointes to Identifier's Value
 */
function calcUnorderedScope(si) {
  var sa = Kit.toArray(si);
  var s = Kit.auto(sa);
  var root = s.first.value;
  function walk(scope) {
    for (var i of s.sub()) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.VariableDeclaration:
            if (i.value.node.kind === "var") {
              Kit.skip(s.peelTo(_core.Tag.declarations));
              var unordered = i.value.node.kind === "var";
              for (var _j of s.sub()) {
                if (_j.enter) {
                  assert.equal(s.cur().pos, _core.Tag.id);
                  for (var _k of s.one()) {
                    if (_k.enter && _k.type === _core.Tag.Identifier && _k.value.sym == null && _k.value.node.name != null) {
                      scope.set(_k.value.node.name, _k.value);
                      _k.value.unordered = unordered;
                    }
                  }
                  if (s.curLev() != null) walk(scope);
                }
              }
              Kit.skip(s.leave());
            }
            break;
          case _core.Tag.Identifier:
            if (i.value.decl) i.value.unordered = null;
            break;
          case _core.Tag.ClassDeclaration:
          case _core.Tag.FunctionDeclaration:
            var j = s.curLev();
            if (j != null) {
              assert.equal(j.pos, _core.Tag.id);
              if (j.value.node.name != null) scope.set(j.value.node.name, j.value);
              j.value.unordered = true;
              walk(i.value.varScope = new Map());
            }
          case _core.Tag.FunctionExpression:
          case _core.Tag.ObjectMethod:
          case _core.Tag.ClassMethod:
            var k = s.curLev();
            if (k != null) {
              var iscope = i.value.varScope = new Map();
              if (k.pos === _core.Tag.id && k.value.node.name != null) iscope.set(k.value.node.name, k.value);
              walk(iscope);
            }
            break;
          case _core.Tag.ArrowFunctionExpression:
            if (s.curLev() != null) walk(i.value.varScope = new Map());
            break;
        }
      }
    }
  }
  walk(root.varScope = new Map());
  return sa;
}

/** puts init field before id in VariableDeclarator and for-of, for-in */
function reorderVarDecl(si) {
  var s = Kit.auto(si);
  function* walk(sw) {
    for (var i of sw) {
      yield i;
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.VariableDeclarator:
            if (i.value.node.kind !== "var") {
              var id = [].concat(_toConsumableArray(walk(s.one())));
              if (s.curLev() != null) {
                assert.equal(s.curLev().pos, _core.Tag.init);
                yield* walk(s.one());
              }
              yield* id;
            }
            break;
          case _core.Tag.AssignmentPattern:
            var left = [].concat(_toConsumableArray(walk(s.one())));
            yield* s.one();
            yield* left;
            break;
          case _core.Tag.ForOfStatement:
          case _core.Tag.ForInStatement:
            var j = s.curLev();
            if (j.type === _core.Tag.VariableDeclaration) {
              var _left = [].concat(_toConsumableArray(walk(s.one())));
              yield* s.one();
              yield* _left;
            }
            break;
        }
      }
    }
  }
  return walk(s);
}

/**
 * assigns unique Symbol object for each variable declaration and usage
 * stores it in sym fields, for root value stores map syms mapping the
 * sym object to a SymbolInfo structure
 * 
 *     interface Sym {
 *       name: String,
 *       orig: String,
 *       num: number,
 *       sym: Symbol,
 *       unordered: boolean,
 *       declScope: TokenValue
 *     }
 * 
 * for each identifier referening variable:
 *
 *     type Value = Value & {sym:Sym,decl?:true}
 * 
 */
var assignSym = exports.assignSym = R.pipe(_core.resetFieldInfo, function assignSym(si) {
  function addVarScope(scope, i) {
    if (i.value.varScope != null) {
      for (var _ref of i.value.varScope) {
        var j = _ref[0];
        var jv = _ref[1];

        scope.set(j, jv.sym || (jv.sym = newSym(j)));
      }
    }
    return scope;
  }
  var sa = Kit.toArray(si);
  var s = Kit.auto(reorderVarDecl(sa));
  var root = s.first.value;
  function decls(scope, unordered, root) {
    for (var i of s.sub()) {
      if (i.enter) {
        switch (i.type) {
          case _core.Tag.Identifier:
            var fi = i.value.fieldInfo;
            if (fi.declVar) {
              i.value.decl = true;
              var _i$value = i.value,
                  name = _i$value.node.name,
                  sym = _i$value.sym;

              if (i.value.unordered) {
                if (sym == null) {
                  sym = i.value.sym = newSym(name);
                  root.varScope.set(name, i.value);
                  scope.set(name, sym);
                }
              } else {
                sym = sym || (i.value.sym = newSym(name));
                scope.set(name, sym);
              }
              if (sym.num == null) sym.num = symNum++;
              sym.name = sym.orig;
              sym.unordered = i.value.unordered || unordered;
              sym.declScope = root;
            } else if (fi.expr || fi.lval) {
              i.value.decl = false;
              if (i.value.sym == null) {
                var _name = i.value.node.name;

                var _sym = scope.get(_name);
                if (_sym == null) {
                  var undef = globals.get(_name);
                  if (undef == null) {
                    _sym = newSym(_name);
                    globals.set(_name, _sym);
                    i.value.sym = _sym;
                    _sym.num = -1;
                    _sym.unordered = false;
                    _sym.declScope = null;
                  } else {
                    i.value.sym = undef;
                  }
                  break;
                }
                i.value.sym = _sym;
              }
            }
            break;
          case _core.Tag.VariableDeclaration:
            decls(scope, i.value.node.kind === "var", root);
            break;
          case _core.Tag.BlockStatement:
          case _core.Tag.ForStatement:
          case _core.Tag.ForInStatement:
          case _core.Tag.ForOfStatement:
            decls(new Map(scope), false, root);
            break;
          case _core.Tag.File:
          case _core.Tag.FunctionDeclaration:
          case _core.Tag.ObjectMethod:
          case _core.Tag.ClassMethod:
          case _core.Tag.FunctionExpression:
          case _core.Tag.ArrowFunctionExpression:
            decls(addVarScope(new Map(scope), i), false, i.value);
            break;
        }
      }
    }
  }
  decls(new Map(), false, s.first.value);
  return sa;
});

/** 
 * for each variable sets its usages scope (list of functions where the 
 * variable is used except declaration function)
 * 
 *     type Sym = Sym & { refScopes: Set<TokenValue> }
 */
function calcRefScopes(si) {
  var sa = Kit.toArray(si);
  var s = Kit.auto(sa);
  function scope(root) {
    for (var i of s.sub()) {
      if (i.enter) {
        if (i.value.func) {
          scope(i.value);
        } else if (i.type === _core.Tag.Identifier) {
          var _si = i.value.sym;
          if (_si != null && _si.declScope !== root) (_si.refScopes || (_si.refScopes = new Set())).add(root);
        }
      }
    }
  }
  scope(s.first.value);
  return sa;
}

/** 
 * for each block calculates a set of variables referenced in it 
 *
 *    type Value = Value * { varRefs?: Sym[] }
 */
function calcBlockRefs(si) {
  var sa = Kit.toArray(si);
  var s = Kit.auto(sa);
  function scope(value, rootDecls) {
    function block(value, decls, refs) {
      function walk(unordered) {
        for (var i of s.sub()) {
          if (i.enter) {
            switch (i.type) {
              case _core.Tag.VariableDeclaration:
                walk(i.value.node.kind === "var");
                break;
              case _core.Tag.FunctionDeclaration:
              case _core.Tag.ObjectMethod:
              case _core.Tag.ClassMethod:
                var j = s.curLev();
                if (j && j.pos === _core.Tag.id) {
                  rootDecls.add(j.value.sym);
                  Kit.skip(s.one());
                }
                scope(i.value, new Set()).forEach(refs.add, refs);
                break;
              case _core.Tag.ArrowFunctionExpression:
              case _core.Tag.FunctionExpression:
                var nroot = new Set();
                var k = s.curLev();
                if (k && k.pos === _core.Tag.id) {
                  nroot.add(k.value.sym);
                  Kit.skip(s.one());
                }
                scope(i.value, nroot).forEach(refs.add, refs);
                break;
              case _core.Tag.ForStatement:
              case _core.Tag.ForInStatement:
              case _core.Tag.ForOfStatement:
              case _core.Tag.Program:
              case _core.Tag.BlockStatement:
                var nrefs = new Set();
                block(i.value, new Set(), nrefs).forEach(refs.add, refs);
                break;
              case _core.Tag.Identifier:
                if (i.value.sym != null) {
                  if (i.value.decl) (unordered ? rootDecls : decls).add(i.value.sym);
                  refs.add(i.value.sym);
                }
                break;
            }
          }
        }
      }
      walk();
      Kit.skip(s.leave());
      value.varRefs = decls.size ? refs : null;
      return [].concat(_toConsumableArray(refs)).filter(j => !decls.has(j));
    }
    if (value.varScope) {
      for (var i of value.varScope.values()) {
        rootDecls.add(i.sym);
      }
    }
    var res = void 0;
    var params = new Set();
    for (var _i of s.sub()) {
      if (_i.enter) {
        if (_i.type === _core.Tag.Identifier && _i.value.decl != null) {
          if (_i.value.decl) params.add(_i.value.sym);
        } else if (_i.pos === _core.Tag.body || _i.pos === _core.Tag.program) {
          for (var j of params) {
            rootDecls.add(j);
          }
          res = block(_i.value, rootDecls, new Set(rootDecls));
        }
      }
    }
    return res || [];
  }
  scope(s.peel().value, new Set());
  return sa;
}

var nameOpts = ["a", "b", "c", "d", "e", "f", "g", "h", "k", "m", "n", "x", "y", "z"];

function namePos(n, pos) {
  if (pos === 0) return n;
  if (pos === 1) return "_" + n;
  return `${n}${pos - 1}`;
}

/** 
 * after adding some names, there may be some naming conflicts
 * this pass resolves them by looking for same name but different symbols ids
 * and renaming them accordingly
 */
function solve(si) {
  var sa = Kit.toArray(si);
  var root = sa[0].value;
  var refsFrames = []; //: Sym[][]
  var symVals = [].concat(_toConsumableArray(globals.values()));
  var anyPat = new Map();
  for (var i of sa) {
    if (i.enter) {
      if (i.value.varRefs != null) {
        var refs = [].concat(_toConsumableArray(i.value.varRefs)).sort((a, b) => a.num - b.num); //:Sym[]
        if (refs.length) refsFrames.push(refs);
        for (var _i2 of refs) {
          if (_i2.name === "") {
            var ap = anyPat.get(_i2);
            if (ap == null) anyPat.set(_i2, ap = new Set());
            refs.forEach(ap.add, ap);
          }
        }
      }
      if (i.type === _core.Tag.Identifier && i.value.decl === true) symVals.push(i.value.sym);
    }
  }
  for (var _ref2 of [].concat(_toConsumableArray(anyPat)).sort((a, b) => a[0].num - b[0].num)) {
    var s = _ref2[0];
    var is = _ref2[1];

    var ism = new Map();
    for (var _i3 of is) {
      if (_i3.name != null) ism.set(_i3.name, (ism.get(_i3) || 0) + 1);
    }var mn = void 0,
        mv = void 0;
    for (var _i4 of nameOpts) {
      var c = ism.get(_i4);
      if (!c) {
        mv = _i4;
        break;
      }
      if (mn == null || mn > c) {
        mn = c;
        mv = _i4;
      }
    }
    s.name = mv;
  }
  // temporarly extending Sym with
  // type Sym = Sym & {dom:Set<number>,namePos:number}
  var store = new Map(); //: Map<String,Sym[][]>,
  for (var _i5 of symVals) {
    var fi = store.get(_i5.name);
    if (fi == null) store.set(_i5.name, fi = new Set());
  }
  for (var _i6 of refsFrames) {
    var _names = new Set(_i6.map(j => j.name));

    var _loop = function (j) {
      var fi = store.get(j);
      if (fi == null) store.set(j, fi = new Set());
      fi.add(_i6.filter(k => k.name === j));
    };

    for (var j of _names) {
      _loop(j);
    }
  }
  var names = new Set(store.keys());
  up: for (var _ref3 of store) {
    var n = _ref3[0];
    var _refs = _ref3[1];

    for (var _i7 of _refs) {
      if (_i7.length > 1) continue up;
    }store.delete(n);
  }
  for (var _ref4 of store) {
    var _ref5;

    var _n = _ref4[0];
    var _refs2 = _ref4[1];

    var syms = new Set((_ref5 = []).concat.apply(_ref5, _toConsumableArray(_refs2)));
    var osyms = [].concat(_toConsumableArray(syms)).sort((a, b) => a.num - b.num);
    for (var _i8 of osyms) {
      _i8.dom = new Set();
    }for (var _i9 of osyms) {
      var pos = 0,
          nn = void 0;
      for (;; pos++) {
        if (_i9.dom.has(pos)) continue;
        nn = namePos(_n, pos);
        if (pos > 0 && names.has(nn)) continue;
        break;
      }
      _i9.name = nn;
      for (var j of _refs2) {
        if (j.length && j[0] === _i9) {
          j.shift();
          for (var k of j) {
            k.dom.add(pos);
          }
        }
      }
    }
  }
  for (var _i10 of sa) {
    if (_i10.enter && _i10.type === _core.Tag.Identifier && _i10.value.sym != null) {
      _i10.value.node.name = _i10.value.sym.name;
    }
  }
  return sa;
}

var prepare = exports.prepare = R.pipe(calcUnorderedScope, assignSym);

var resolve = exports.resolve = R.pipe(resetSym, calcBlockRefs, solve);