"use strict";

exports.__esModule = true;
exports.all = exports.lazy = exports.BROWSER_DEBUG = undefined;
exports.cg = cg;
exports.ccg = ccg;
exports.verify = verify;
exports.cleanComments = cleanComments;
exports.setSymComments = setSymComments;

var _assert = require("assert");

var assert = _interopRequireWildcard(_assert);

var _babelGenerator = require("babel-generator");

var _babelGenerator2 = _interopRequireDefault(_babelGenerator);

var _babelTypes = require("babel-types");

var T = _interopRequireWildcard(_babelTypes);

var _core = require("./core");

var _chalk = require("chalk");

var _chalk2 = _interopRequireDefault(_chalk);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

var MAX_TRACE_CODE_LEN = 40;
var TYPE_SIZE = 20;
var BROWSER_DEBUG = exports.BROWSER_DEBUG = typeof window !== "undefined" && window.chrome;
var enterStyles = "background: #222; color: #bada55;font-size:1.5em";
var leaveStyles = "color: #ee5757; background: black";
var newTagStyle = "color:purple;font-size:large";
var positionStyle = "color:gray;text-decoration:underline";
var dirStyle = "font-size:x-large;font-weight:bolder";

/**
  * outputs short description of AST node
  */
function cg(ast) {
  var opts = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var res;
  if (ast == null) {
    // console.error("<NULL>")
    return "<NULL>";
  }
  try {
    if (Array.isArray(ast)) {
      ast = ast.length > 0 && T.isExpression(ast[0]) ? T.sequenceExpression(ast) : T.blockStatement(ast);
    }
    res = (0, _babelGenerator2.default)(ast, opts, "").code;
  } catch (e) {
    if (ast.type != null) console.error(e.stack);
  }
  if (res != null) {
    return res;
  }
  return "!!" + opts.compact ? JSON.stringify(ast) : JSON.stringify(ast, null, 2);
}

/**
  * same as `cg` but using compact syntax
  */
function ccg(ast) {
  return cg(ast, { compact: true });
}

if (BROWSER_DEBUG) {
  global.cg = cg;
  global.ccg = ccg;
}

function* verify(s) {
  var stack = [];
  for (var i of (0, _core.resetFieldInfo)(s)) {
    assert.ok(i.enter != null);
    assert.ok(i.leave != null);
    assert.ok(i.pos != null);
    assert.ok(i.type != null);
    assert.ok(i.value != null);
    var ti = (0, _core.typeInfo)(i);
    var ctrlPos = (0, _core.symInfo)(i.pos).kind === "ctrl";
    if (i.enter && stack.length) {
      var _stack = stack[stack.length - 1],
          f = _stack[0],
          keys = _stack[1];

      if (f.type === _core.Tag.Array) {
        if (!ctrlPos) assert.equal(i.pos, _core.Tag.push);
      } else if (keys != null && !ctrlPos) {
        var k = void 0;
        while ((k = keys.shift()) != null) {
          if (_core.Tag[k] === i.pos) break;
        }
        assert.ok(k, "field name exists");
      }
    }
    if (i.enter && i.value.fieldInfo != null) {
      if (i.type === _core.Tag.Array) {
        assert.ok(i.value.fieldInfo.array, "expected array field");
      } else if (ti.kind === "node") {}
    }
    if (i.enter && !i.leave) {
      var _keys = T.VISITOR_KEYS[(0, _core.symName)(i.type)];
      stack.push([i, _keys && [].concat(_toConsumableArray(_keys))]);
    }
    if (!i.enter && i.leave) {
      var _stack$pop = stack.pop(),
          _f = _stack$pop[0];

      assert.ok(_f != null);
      if (ti.kind !== "ctrl") assert.equal(_f.type, i.type);
      assert.equal(_f.pos, i.pos);
      assert.equal(_f.value, i.value);
    }
    yield i;
  }
  assert.equal(stack.length, 0);
}

function pad(s) {
  if (s.length % 2) s += ' ';
  var sps = Array(Math.max(20 - s.length / 2, 2)).join(' ');
  return sps + s + sps;
}

var traceImpl = BROWSER_DEBUG ? browserTraceImpl : traceNodeImpl;

function* traceNodeImpl(prefix, s) {
  var level = 0;
  var x = 0;
  //  prefix = chalk.bold(prefix)
  for (var i of s) {
    if (i.enter) level++;
    var dir = _chalk2.default.bold(i.leave ? i.enter ? "|" : "/" : "\\");

    var clevel = s.level ? `/${s.level}` : "";
    var descr = `${_chalk2.default.green((0, _core.symName)(i.pos))}:${i.type.ctrl ? _chalk2.default.green((0, _core.symName)(i.type)) : _chalk2.default.green.bold((0, _core.symName)(i.type))}[${level}${clevel}]`;
    var n = "";
    var node = i.value.node;

    var comments = [];
    var commentsTxt = "";
    var t = [].concat(i.value.comments || [], i.value.tcomments || []);
    if (t.length) {
      for (var j of t) {
        var c = j.txt;
        c = !i.enter(_chalk2.default.dim(j.txt)) || j.txt;
      }
      if (comments.length) commentsTxt = _chalk2.default.bold("[") + comments.join(" ") + _chalk2.default.bold("]");
    }
    if (node != null && i.type !== _core.Tag.Array && (0, _core.symKind)(i.type) !== "ctrl") {
      n = ccg(node);
      if (n.length > MAX_TRACE_CODE_LEN) n = n.substr(0, MAX_TRACE_CODE_LEN) + "...";
      n = _chalk2.default.yellow(n);
      var loc = node.loc;

      if (loc != null) {
        var f = loc.source,
            _s = loc.start,
            e = loc.end;

        n += _chalk2.default.blue(` @${f || "?"}-${_s.line}:${_s.column}..${e.line}:${e.column}`);
      } else {
        n += _chalk2.default.bold(" @new");
      }
    }
    console.log(prefix, Array(level).join(' '), dir, `${descr}@${x}`, commentsTxt, n);
    yield i;
    if (i.leave) {
      level--;
    }
    x++;
  }
  console.log(prefix, "len:", x);
}

function* browserTraceImpl(prefix, s) {
  var level = 0;
  var x = 0;
  console.log(`%c${pad(prefix)}%c`, `background:#2B81AF;color:#fff;font-size:xx-large;
              text-shadow:rgba(0, 0, 0, 0.5) 2px 2px 1px`, "");
  for (var i of s) {
    var _console;

    var styles = [];
    if (i.enter) level++;
    var dir = i.enter && i.leave ? "\u21c4" : i.enter ? "\u2192" : "\u2190";
    var clevel = s.level ? `/${s.level}` : "";
    var tp = i.pos === i.type ? (0, _core.symName)(i.type) : `${(0, _core.symName)(i.pos)}:${(0, _core.symName)(i.type)}`;
    var descr = `${prefix}${tp}[${level}${clevel}]`;
    var ldescr = void 0;
    if (i.pos === i.type) {
      ldescr = `${prefix}%c${(0, _core.symName)(i.type)}%c[${level}${clevel}]`;
      styles.push((0, _core.symKind)(i.type) === "ctrl" ? "color:navy;font-size:large" : "color:steelblue", "");
    } else {
      ldescr = `${prefix}%c${(0, _core.symName)(i.pos)}:%c${(0, _core.symName)(i.type)}%c[${level}${clevel}]`;
      styles.push((0, _core.symKind)(i.pos) === "ctrl" ? "color:olive;font-size:large" : "color:green");
      styles.push((0, _core.symKind)(i.type) === "ctrl" ? "color:navy;font-size:large" : "color:steelblue", "");
    }
    if (i.enter && !i.leave && console.group != null) {
      if (x === 0 && i.value.collapsed !== false || i.value.collapsed) console.groupCollapsed(descr);else console.group(descr);
    }
    var n = "";
    var node = i.value.node;

    var comments = [];
    var commentsTxt = "";
    var commentsStyle = [];
    var t = [].concat(i.value.comments || [], i.value.tcomments || []);
    if (t.length) {
      var pstyle = "color:green;font-size:large";
      styles.push(pstyle);
      for (var j of t) {
        comments.push(`%c${j.txt}%c`);
        var mod = !i.enter ? "font-size:small;font-style:italic" : "font-weight:bolder;font-size:large";
        var _s2 = `${j.style || ''}${mod}`;
        styles.push(_s2, "");
      }
      if (comments.length) {
        commentsTxt = "%c[" + comments.join(" ") + "%c]%c";
        styles.push(pstyle, "");
      }
    }
    if (node != null && i.type !== _core.Tag.Array && (0, _core.symKind)(i.type) !== "ctrl") {
      n = ccg(node);
      if (n.length > MAX_TRACE_CODE_LEN) n = n.substr(0, MAX_TRACE_CODE_LEN) + "...";
      n = `%c ${n} %c`;
      styles.push(i.enter ? enterStyles : leaveStyles, "");
      var loc = node.loc;

      if (loc != null) {
        var f = loc.source,
            _s3 = loc.start,
            e = loc.end;

        n += ` %c@${f || "?"}-${_s3.line}:${_s3.column}..${e.line}:${e.column}%c`;
        styles.push(positionStyle, "");
      } else {
        n += " %c@new%c";
        styles.push(newTagStyle, "");
      }
    }
    (_console = console).log.apply(_console, [`%c${dir}%c ${ldescr}@${x}${commentsTxt} ${n}`, dirStyle, ""].concat(styles, [i.value]));
    yield i;
    if (i.leave) {
      if (!i.enter && console.group != null) console.groupEnd();
      level--;
    }
    x++;
  }
  console.log(`${prefix}: len: ${x}`);
}

function traceAllImpl(prefix, s) {
  return [].concat(_toConsumableArray(verify(traceImpl(prefix, s))));
}

function traceArgs(impl) {
  return function traceImpl(prefix, s) {
    if (prefix == null || prefix.substr == null) {
      if (s == null) s = prefix;
      prefix = "";
    }
    if (prefix.length) prefix += ":";
    if (s == null || s[Symbol.iterator] == null) return s => impl(prefix, s);
    return impl(prefix, s);
  };
}

function* cleanComments(s) {
  for (var i of s) {
    if (i.enter) i.value.comments = [];
    yield i;
  }
}

function* setSymComments(s) {
  for (var i of s) {
    if (i.enter && i.value.sym && i.value.sym.id) {
      var c = i.value.comments || (i.value.comments = []);
      c.push({ txt: i.value.sym.id });
    }
    yield i;
  }
}

var lazy = exports.lazy = traceArgs(traceImpl);
var all = exports.all = traceArgs(traceAllImpl);
exports.default = lazy;